#p_win {
}

#c_win.black, #c_frame.black {
    background-color: #000;
}

#c_win.white , #c_frame.white {
    background-color: #fff;
}

#annotations_textview {
  font-size: 1.4em;
}

#p_win GtkPaned .pane-separator {
    background-color: #ccc;
    border-radius: 2px;
}

#p_win GtkFrame {
}

#p_win GtkTreeView {
    border-style: none;
    background-color: transparent;
    font-size: 16pt;
}

#label_time, #label_ett, #frame_clock {
  font-family: "monospace";
}

#frame_ett {
  /* similar to button padding to add some room */
  margin-left: 3px;
  margin-right: 3px;
}
#frame_cur {
  min-width: 15em;
}
#frame_ett {
  min-width: 6em;
}
#frame_clock {
  min-width: 8em;
}

.toolbar {
  background: @theme_base_color
}

#scribble_overlay {
  background: @theme_bg_color
}

#scribble_overlay modelbutton.pen-preset radio {
  margin: 0 30px 25px 0;
}

.frame-label {
    font-size:1.2em;
}

.big-info-label {
    font-size: 1.6em;
}

.info-label {
    font-size: 1.4em;
    opacity: 0.80;
}

.big-info-label > .frame-label {
    font-size:0.7em;
}

.info-label > .frame-label {
    font-size:0.8em;
}

.ett-reached {
    color: @success_color;
}
.ett-info {
    color: @warning_color;
}
.ett-warn {
    color: @error_color;
}
.time-warn {
    animation: blinker 250ms cubic-bezier(.5, 0, 1, 1) infinite alternate;
}
@keyframes blinker {
  from { opacity: 1; }
  to { opacity: 0; }
}

.grid-frame > *:hover {
  border-color: @theme_selected_bg_color;
}
